/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: muaddib_atreides@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "nSX2.h"
#include "elf_file.h"

FILE * ElfFile;
elf_header ico_elf_hdr;
elf_program_header ico_elf_prg_hdr[30];
elf_section_header ico_elf_sec_hdr[30];
long ElfFileSize;

FILE * BiosFile;


long FileSizeOf( FILE *fp )
{
    long int save_pos;
    long size_of_file;

    save_pos = ftell( fp );
    fseek( fp, 0L, SEEK_END );
    size_of_file = ftell( fp );
    fseek( fp, save_pos, SEEK_SET );
    return( size_of_file );
}


_u8 ElfOpenFile( char * fFileName ) {

	_u8 	Result;

	 int errvalue;
	
	ElfFile = fopen( fFileName, "r" );
	if( ElfFile != NULL )  {
		ElfFileSize = FileSizeOf( ElfFile ) ;
		printf("File:Opening Elf : %s ( size : %5.3f Ko)  \n" ,fFileName , (float)( ElfFileSize  / 1024) );
		Result = TRUE; 
	}
	else	{
		printf("File:Error opening :%s: ", fFileName );
		errvalue = errno;
//		printf( "The error generated was %d\n", errvalue );
  	 	printf( ": %s\n", strerror( errvalue ) );
		
		
		Result = FALSE;
	}
	
	return Result;	
}

_u8 BiosOpenFile( char * fFileName ) {

	_u8 	Result;
	long BiosFileSize;
	
	BiosFile = fopen( fFileName, "r" );
	if( BiosFile != NULL )  {
		BiosFileSize = FileSizeOf( BiosFile ) ;
		printf("File:Opening BIOS : %s ( size : %5.3f Ko)  \n" ,fFileName , (float)( BiosFileSize  / 1024) );
		Result = TRUE; 
	}
	else	{
		printf("File:Error opening %s \n", fFileName );
		Result = FALSE;
	}
	
	return Result;	
}


void ElfCloseFile( void ) {
		printf("File:Close Elf File\n");
		fclose( ElfFile );
}

void BiosCloseFile( void ) {
		printf("File:Close Bios File\n");
		fclose( BiosFile );
}

_u8 ElfLoadHeader() {
	
	_u8  Index;
	_u8 Result;
	
	Result = 0;
	
	fseek( ElfFile, 0L, SEEK_SET );
	if ( fread( &ico_elf_hdr , sizeof(struct  _elf_header ), 1, ElfFile ) != 0 ) {
		printf("File:Read  Elf Header\n");
		MipsCpu.EntryPointAdd = ico_elf_hdr.entry_point_add;
		}
	else Result++;
	
	fseek( ElfFile, ico_elf_hdr.start_prg_hdr, SEEK_SET );
	printf("File:Read  Elf Program Header Numero 0 to");
	for ( Index = 0 ; Index <  ico_elf_hdr.num_of_prg_hdr ; Index++ ) {
		
		if ( fread( &ico_elf_prg_hdr[Index] , sizeof(struct  _elf_program_header ), 1, ElfFile ) == 0 ) Result++;
		
		//fseek( ElfFile, ico_elf_hdr.start_prg_hdr+ ( sizeof(struct  _elf_program_header ) * (Index + 1) ) +ico_elf_prg_hdr[Index].num_of_bytes_in_file_ima , SEEK_SET );
	}
	printf(" %d\n",Index-1);
		
	fseek( ElfFile,  ico_elf_hdr.start_sct_hdr, SEEK_SET );
	printf("File:Read  Elf Section Header Numero 0 to");
	for ( Index = 0 ; Index <  ico_elf_hdr.num_of_sec_hdr ; Index++ ) {
		if ( fread( &ico_elf_sec_hdr[Index] , sizeof(struct  _elf_section_header ), 1, ElfFile ) == 0 )	Result++;
			
	}	
	printf(" %d\n",Index-1);

	if ( Result == 0 ) 		printf("File:Read  Elf Header DONE\n");
	else printf("File:Read  Elf Header ERROR");

	return Result;
}



void ElfPrintInfo() {
	_u8	index;

	printf("--------------------------------------------------------------------------\n");
	printf("ELF INFO\n");
	printf("Entry Point Adresse : %lx\n", ico_elf_hdr.entry_point_add );
	printf("\n");	
	printf("Start Of Program Header : %lx\n", ico_elf_hdr.start_prg_hdr );
	printf("Number Of  Program Header : %x\n", ico_elf_hdr.num_of_prg_hdr );
	for( index = 0 ; index < ico_elf_hdr.num_of_prg_hdr; index++) {
		printf("ProgramHeader %d\n",index);
		printf(" Offset from file start to program segment : %lx\n", ico_elf_prg_hdr[index].offs_from_start_to_prg_seg );
		printf(" Virtual Adress of Segement : %lx\n", ico_elf_prg_hdr[index].virtual_add_of_segment );
		printf(" Physical Adress of Segment  : %lx\n", ico_elf_prg_hdr[index].physic_add_of_segment );
		printf(" Number of bytes in file image  : %lx ( %ld ) \n", ico_elf_prg_hdr[index].num_of_bytes_in_file_ima, ico_elf_prg_hdr[index].num_of_bytes_in_file_ima );
		printf(" Number of bytes in mem image : %lx ( %ld ) \n", ico_elf_prg_hdr[index].num_of_bytes_in_mem_ima, ico_elf_prg_hdr[index].num_of_bytes_in_mem_ima );
	}
	printf("\n");
	printf("Start Of Section Header : %lx ( %ld )\n", ico_elf_hdr.start_sct_hdr, ico_elf_hdr.start_sct_hdr  );
	printf("Number of  Section Header : %x ( %d )\n", ico_elf_hdr.num_of_sec_hdr ,  ico_elf_hdr.num_of_sec_hdr );
	printf("Section Header Entry Size: %x ( %d )\n", ico_elf_hdr.sec_hdr_entry_size,  ico_elf_hdr.sec_hdr_entry_size);
	printf("Section Header Stringatable Index: %x ( %d )\n", ico_elf_hdr.sec_hdr_stringa_index, ico_elf_hdr.sec_hdr_stringa_index );
	printf("\n");
	
	for( index = 0 ; index < ico_elf_hdr.num_of_sec_hdr; index++) {
		ElfPrintSectionInfo( index );
		getchar();
	}	
		
	printf("--------------------------------------------------------------------------\n");
}


void ElfPrintSectionInfo( _u8 fIndex ) {
	printf("--------------------------------------------------------------------------\n");

	printf("Section numero : %d\n",fIndex);
	printf(" Type : %lx ( %ld )\n", ico_elf_sec_hdr[fIndex].type, ico_elf_sec_hdr[fIndex].type );
	printf(" Numero Index Section : %lx ( %ld )\n", ico_elf_sec_hdr[fIndex].numero_index_sect,  ico_elf_sec_hdr[fIndex].numero_index_sect );
	printf(" Section Start Adress : %lx ( %ld )\n", ico_elf_sec_hdr[fIndex].sect_start_add, ico_elf_sec_hdr[fIndex].sect_start_add );
	printf(" Offset from start : %lx ( %ld )\n", ico_elf_sec_hdr[fIndex].offset_from_start ,  ico_elf_sec_hdr[fIndex].offset_from_start  );
	printf(" Size : %lx ( %ld ) \n", ico_elf_sec_hdr[fIndex].size,  ico_elf_sec_hdr[fIndex].size  );
	printf(" Alignement : %lx ( %ld ) \n", ico_elf_sec_hdr[fIndex].alignement,  ico_elf_sec_hdr[fIndex].alignement  );
	printf(" Fixed Size Entry : %lx ( %ld ) \n", ico_elf_sec_hdr[fIndex].fixed_size_entry, ico_elf_sec_hdr[fIndex].fixed_size_entry);

	printf("--------------------------------------------------------------------------\n");
}