/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: muaddib_atreides@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <SDL/SDL.h>


#include "nSX2.h"
#include "plugin_gs/PS2Edefs.h"


Mips_CPU 					MipsCpu;
Bios_Hw  					Ps2Bios;

SDL_Surface		* 		MainScreen;
SDL_Event		*		MainSDLEvent;

static int filter(const SDL_Event *event)
{
   switch (event->type)
     {
      case SDL_KEYDOWN:
	switch (event->key.keysym.sym)
	  {
	   case SDLK_ESCAPE: MipsCpu.halt = 1;   break;
	   case SDLK_F1:     break;
	   default:
	    
	  }
	return 0;
	break;
      case SDL_KEYUP:
	switch (event->key.keysym.sym)
	  {
	   case SDLK_ESCAPE:     break;
	   case SDLK_F1:     break;
	   default:
	    
	  }
	return 0;
	break;
      default:
	return 1;
     }
}






char ElfToExecute[256];



int main( void ) {
	
	int 	result;
	char GSMainTitle[256];
	
	printf("nSX2 %d.%d by Muad - 2002 \n",  nSX2_VERSION, nSX2_BUILD);

	/* Read config File */
	result = nSX2LoadConfig();												

	/*Open the elf File */
	result = ElfOpenFile(ElfToExecute);					if ( result == FALSE )  return TRUE;
	
	/*Open the BIOS File */
	result = BiosOpenFile( "bios/scph10000.bin" ) ;		if ( result == FALSE )  return TRUE;
	
	/*Various Memory Allocation */
	result = MipsMemoryAlloc();							if ( result == FALSE )  return TRUE;
	result =MipsSpecialAreaMemoryAlloc( );				if ( result == FALSE )  return TRUE;
	result =MipsGSAreaMemoryAlloc();					if ( result == FALSE )  return TRUE;
	result =MipsBiosMemoryAlloc( );						if ( result == FALSE )  return TRUE;

	/*Read the header of the Elf File*/
	ElfLoadHeader();

	/* */
	//ElfPrintInfo();
	
	

	/*Copy the program to memory according the elf header*/
	CopyProgramToMemory(  ) ;
	
	/*Copy data to memory according the elf header*/
	CopySectionToMemory( );
	
	/*Copy Bios to Memory */
	CopyBiosToMemory ( );

	
	
	/* Initialisation of GS plugin */
	sprintf(GSMainTitle, " %s %d.%d - %s", "nSX2", nSX2_VERSION, nSX2_BUILD,ElfToExecute);
	GSinit();
	GSopen(MainScreen ,GSMainTitle );
	
	/* SDL Filter Event */
	SDL_SetEventFilter(filter);
	
	
	/*Various PS2 Initialisation */
	Ps2Bios.NeedUpdateScreen = 0x6FFF;
	MipsCpu.PC = MipsCpu.EntryPointAdd ;				//Start Program At Addr
	MipsCpu.GPR[28].lo	= 0x81F80000;					//GP
	MipsCpu.GPR[29].lo	= 0x81F00000;					//SP
	MipsCpu.CPR0[12][0]	=	0x10900001;			//Reg Status of COP0
	MipsCpu.CPR0[15][0]	=	0x2E20	;					//Reg PrID of COP0
	
	printf("\nStart Interpreter...\n\n");
	do {

		//if ( MipsCpu.PC == 0x00201a90 ) debug_op = 1;

		/*Execute the current opcode*/
		MipsUpdateCpu ( );

		/*Execute DMA stuff*/
		DmaUpdate();
		
		/*Execute GS stuff*/
		GSUpdate();
		
		if ( debug_op == 1 )  {
			 MipsCpuPrintReg(  );
			 getchar();
		 }
		
		SDL_PollEvent( MainSDLEvent);
	
	} while( MipsCpu.halt != 1 );	
	
	printf("\nStop Interpreter...\n\n");	
	
	printf("Stop at PC = %llx\n",MipsCpu.PC);
	PrintDebugOpcode() ;
	MipsCpuPrintReg(  );
	
	/* Ending */
	ElfCloseFile();
	BiosCloseFile( ) ;
	
	/*Free All Memory*/
	IcoFreeMemory();
	
	/*Free GS plugin Memory*/
	GSclose();

	/*All done...*/	
	return TRUE;
	
}
