/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: muaddib_atreides@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#define nSX2_VERSION	0
#define nSX2_BUILD			1

#define GSdefs

#define TRUE 	1
#define FALSE 	0

typedef unsigned long long _u64;
typedef unsigned long _u32;
typedef unsigned short _u16;
typedef unsigned char _u8;

typedef struct __u128 {
	_u64 lo;
	_u64 hi;
} _u128;



typedef	struct _Mips_CPU {
	_u32		opcode;
	_u8		halt;
	
	_u128 		GPR[34];				//GENERAL 	REGISTER
	_u64		CPR0[32][4];			//COP0			REGISTER
	_u128		FPR1[32];				//COP1	FP		REGISTER
	
	_u64		PC;
	_u64 	EntryPointAdd;
	
} Mips_CPU;




typedef struct _DmaChannel {

	_u64		ctrl;
	_u64		memAdd;
	_u64		size;
	_u64		tag;
	_u64		asr0;
	_u64		asr1;
	_u64		saddr;

} DmaChannel;


typedef	struct _Bios_Hw {

	_u64		GsCrt;
	
	/*
	Permet de detecter les acces au registre du GS
	*/
	_u8		WriteToGSArea;
	_u64		GSAreaWriteAt;
	_u64		GSAreaWriteVal;

	/*
	Permet de detecter les acces au registre DMA
	*/
	_u8		WriteToDmaArea;
	_u64	DmaAreaWriteAt;
	_u64	DmaAreaWriteVal;
	
	_u8		ReadToDmaArea;
	_u64	DmaAreaReadAt;
	_u64	DmaAreaReadVal;	
	
	/* DMA CHANNEL */
	DmaChannel DmaC[10];
	
	_u64	NeedUpdateScreen;

} Bios_Hw;





extern FILE * ElfFile;
extern long ElfFileSize;
extern FILE * BiosFile;


extern Mips_CPU MipsCpu;
extern Bios_Hw  Ps2Bios;



extern _u8 *MipsMemory;
extern _u8	*SpecialAreaMemory;
extern _u8	*GSAreaMemory;
extern _u8   *BiosMemory;

extern char ElfToExecute[256];


#define sign_extended(a) if (a & 0x0000000080000000LL) a |= 0xFFFFFFFF00000000LL; else a &= 0xFFFFFFFF
#define sign_extendedb(a) if (a & 0x80) a |= 0xFFFFFFFFFFFFFF00LL; else a &= 0xFF
#define sign_extendedh(a) if (a & 0x8000) a |= 0xFFFFFFFFFFFF0000LL; else a &= 0xFFFF


//This macro decode the memory and Fill the opcode field of the cpu
#define MipsFillOpcode(A) MipsCpu.opcode=(MipsMemory[A+3]<<24)+(MipsMemory[A+2]<<16)+(MipsMemory[A+1]<<8)+(MipsMemory[A])

//These macro Extract bits from opcode and fill appropriate var
#define Extract_rt() rt=(MipsCpu.opcode&0x1F0000)>>16
#define Extract_ft() ft=(MipsCpu.opcode&0x1F0000)>>16

#define Extract_rs() rs=(MipsCpu.opcode&0x3E00000)>>21
#define Extract_base() base=(MipsCpu.opcode&0x3E00000)>>21

#define Extract_imm() imm=(MipsCpu.opcode&0xFFFF)
#define Extract_offset() offset=(MipsCpu.opcode&0xFFFF)

#define Extract_rd() rd=(MipsCpu.opcode&0xF800)>>11

#define Extract_instr_index() instr_index=(MipsCpu.opcode&0x3FFFFFF)
#define Extract_sa() sa=(MipsCpu.opcode&0x7C0)>>6
#define Extract_sel() sa=(MipsCpu.opcode&0x7)


//This macro decode the memory and a var named B
#define ReadMemAt(B,A) B=(MipsMemory[A+3]<<24)+(MipsMemory[A+2]<<16)+(MipsMemory[A+1]<<8)+(MipsMemory[A])


/*
ELF FILE
*/
long FileSizeOf( FILE *fp );
_u8 ElfOpenFile( char * fFileName );
void ElfCloseFile( void );
_u8 ElfLoadHeader() ;
void ElfPrintInfo() ;
void ElfPrintSectionInfo( _u8 fIndex ) ;

void BiosCloseFile( void ) ;
_u8 BiosOpenFile( char * fFileName ) ;

/*
R5900 MEMORY
*/
#define DATA_QUAD		1		//128
#define DATA_DW			2		//64
#define DATA_DWH		32		//64
#define DATA_W			4		//32
#define DATA_HW			8		//16
#define DATA_B			16		//8

_u8 MipsMemoryAlloc( void );
_u8 MipsSpecialAreaMemoryAlloc( void );
_u8 MipsGSAreaMemoryAlloc( void )  ;
_u8 MipsBiosMemoryAlloc( void )  ;

void IcoFreeMemory( void ) ;
void CopyProgramToMemory( void ) ;
void CopySectionToMemory ( void ) ;
void CopyBiosToMemory (  void ) ;

void	MipsMemoryWrite ( _u128 fVal  , _u8 fDataType , _u64 fBase , int fOffset ) ;
void MipsSpecialMemWrite( _u8 * Mem, _u128  fVal  , _u8 fDataType , _u64 fBase , int fOffset );
void MipsGSAreaMemWrite( _u8 * Mem , _u128  fVal  , _u8 fDataType , _u64 fBase , int fOffset ) ;
void	MipsMemoryWriteDW (_u8 * Mem, _u64 fVal , _u64 fAdd ) ;
void MipsMemoryWriteW  (_u8 * Mem, _u64 fVal , _u64 fAdd ) ;
void MipsMemoryWriteHW  (_u8 * Mem, _u64 fVal , _u64 fAdd ) ;
void MipsMemoryWriteB  (_u8 * Mem, _u64 fVal , _u64 fAdd ) ;

void	MipsMemoryRead ( _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) ;
void MipsSpecialMemRead(_u8 * Mem, _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) ;
void MipsGSAreaMemRead(_u8 * Mem, _u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) ;
void  MipsBiosAreaMemRead(_u8 * Mem,_u128 * fVal  , _u8 fDataType , _u64 fBase , int fOffset ) ;

void MipsMemoryReadDWH (_u8 * Mem, _u128 * fVal , _u64 PhysicalAdd );
void MipsMemoryReadDW (_u8 * Mem, _u128 * fVal , _u64 PhysicalAdd );
void MipsMemoryReadW (_u8 * Mem, _u128 * fVal , _u64 fAddr );
void MipsMemoryReadHW (_u8 * Mem, _u128 * fVal , _u64 fAddr ) ;
void MipsMemoryReadB ( _u8 * Mem,_u128 * fVal , _u64 fAddr ) ;


/* 
CONFIG FILE 
*/

int nSX2LoadConfig ( void ) ;






/*
R5900 CPU
*/

void MipsUpdateCpu ( ) ;
void	PrintDebugOpcode() ;
void MipsCpuPrintReg( void ) ;
_u8 	debug_op	;
_u8  debug_gs;
_u8 debug_dma;
_u8 debug_print_syscall;

/*
GS MAIN
*/
void GSUpdate() ;




/*
DMA MAIN
*/
void		DmaUpdate() ;
